# ~/ryu$ PYTHONPATH= ./bin/ryu-manager ryu/app/cpm_13.py --observe-links

from ryu.ofproto import ofproto_v1_3
from ryu.app import simple_switch_13
from ryu.controller.handler import set_ev_cls
from ryu.topology import event, switches
from ryu.topology.api import get_switch, get_link
from ryu.controller import ofp_event
from ryu.controller.handler import CONFIG_DISPATCHER, MAIN_DISPATCHER
from ryu.lib.packet import packet
from ryu.lib.packet import arp
from ryu.lib.packet import ethernet
from ryu.lib.packet import ether_types
from ryu.lib import hub
import time
from my_http_client import MyHTTPClient


class CPM13(simple_switch_13.SimpleSwitch13):
    "CPM module for topology discovery and SPF"

    OFP_VERSIONS = [ofproto_v1_3.OFP_VERSION]

    def __init__(self, *args, **kwargs):
        super(CPM13, self).__init__(*args, **kwargs)
        self.topo_links = {}
        self.switches = {}
        self.neighbor_ports = {}  # Temp 2017-10-02
        self.port_metric = {}  # 2017-10-16
        self.hosts = {}
        self.hosts_temp_ban = {}  # 2017-10-17
        self.sample_interval = 30  # How often to query the DM
        self.ref_bw = 100000000.0  # 100 Mbps; Reference Bandwidth in Bps
        self.ref_hc = 30.0  # 30 Hops; Reference Hop Count
        # Connect with DM:
        self.dm_server = MyHTTPClient('http://192.16.125.183/API_REST/request.php/get?')
        self.port_counters_snapshot = {'object_index': 1000, 'switches': []}
        self.get_dm_data_thread = hub.spawn(self._get_dm_data)

    def _get_dm_data(self):
        while True:
            # Ask for DM data every self.sample_interval seconds:
            hub.sleep(self.sample_interval)
            #body = self.dm_server.get_data_from_dm('port_counters', 1)
            params_dict = {'module': 'nfm', 'row': 1, 'table': 'port_counters'}
            body = self.dm_server.send_GET(params_dict)
            #print body
            if body is None:  # This is True if NO json in HTTP reponse
                print 'No NFM monitoring data received from DM'
                continue
            # print body
            self._calculate_link_metric(body)

    def _calculate_link_metric(self, data_dict):

        for switch in data_dict['switches']:
            # each element is a dictionary as well
            for port_counter in switch['port_counters']:
                # each element is a dictionary as well

                tx_metric = float(port_counter['tx_bw']) / self.ref_bw

                self.port_metric[(switch['switch_dpid'],
                                 port_counter['port'])] = tx_metric

                #print '[CALC] Port Metric', (switch['switch_dpid'], port_counter['port']), 'set to', tx_metric

    def _calculate_hop_count(self, paths):
        paths_dict = {}

        for path in paths:
            hop_count = len(path)
            paths_dict[path] = hop_count / self.ref_hc

        return paths_dict

    def _calculate_path_metric(self, paths):
        paths_dict = {}

        for path in paths:
            # Find the busiest port in the path.
            # Path metric is equal to the busiest port's metric.
            worst_metric = 0
            for k, v in enumerate(path[:-1]):
                # v is the current switch dpid
                # path[k + 1] is the dpid of the next switch in path
                (tx_port, _) = self.neighbor_ports[(v, path[k + 1])]
                # Find busiest port in path:
                if worst_metric < self.port_metric[(v, tx_port)]:
                    worst_metric = self.port_metric[(v, tx_port)]
            paths_dict[path] = worst_metric

        return paths_dict

    @set_ev_cls(event.EventSwitchEnter)
    def _handle_new_switch(self, ev):
        # Gives the ryu.controller.controller.Datapath object of the switch:
        # print ev.switch.dp
        new_switch = ev.switch.dp
        self.switches[new_switch.id] = ev.switch.dp
        self.topo_links[new_switch.id] = []

    @set_ev_cls(event.EventLinkAdd)
    def _handle_new_link(self, ev):
        # The link obviously connects two switches. The following two:
        # SwitchA: ev.link.src
        # SwitchB: ev.link.dst
        # are Port objects (switches.py)
        # print 'New link between', ev.link.src.dpid, 'and', ev.link.dst.dpid

        switchA = ev.link.src
        switchB = ev.link.dst

        # print switchA

        self.topo_links[switchA.dpid].append(switchB.dpid)
        # I don't need to report the other direction.
        # SwitchB will also raise this event.

        # {11: [12, 21], 12: [11, 13], 13: [22, 12], 22: [21, 13],
        # 21: [22, 11]}

        # print 'Neighbors', self.topo_links

        # Temp 2017-10-02
        self.neighbor_ports[(switchA.dpid, switchB.dpid)] = (
            switchA.port_no, switchB.port_no)
        # print 'Ports', self.neighbor_ports

        # 2017-10-16
        self.port_metric[(switchA.dpid, switchA.port_no)] = 0
        #print 'Port Metric', (switchA.dpid, switchA.port_no), 'set to', 0

    # https://www.python.org/doc/essays/graphs/
    def _find_all_paths(self, neighbors, start, end, path=[]):
        path = path + [start]
        if start == end:
            return [path]
        if start not in neighbors.keys():
            return []
        paths = []
        for node in neighbors[start]:
            if node not in path:
                newpaths = self._find_all_paths(neighbors, node, end, path)
                for newpath in newpaths:
                    paths.append(newpath)
        return paths

    def _find_paths(self, neighbors, start, end):
        # 2017-10-19
        # Change to list of tuples, instead of list of lists

        if start == end:
            # same switch dpid i.e. the hosts are connected on the same switch
            # So skip path calculation and return None
            return None
        else:
            paths = self._find_all_paths(neighbors, start, end)
            temp_paths = []
            for j in paths:
                temp_paths.append(tuple(j))
            return temp_paths

    def _add_end_to_end_flow(self, path, src_host_ip, dst_host_ip):
        # input argument "path" is a list of switches; each item is a dpid
        # First I need to find the ports of the two switches:
        if len(path) == 1:
            print 'Hosts are on same switch', path[0]
            return

        hostA = self.hosts[src_host_ip]  # src_host
        hostZ = self.hosts[dst_host_ip]  # dst_host

        in_port = hostA['port']

        # Installing flows from A >> Z
        # I am ommiting last switch as _handle_new_host()
        # will add a flow for hostZ there.
        for k, v in enumerate(path[:-1]):
            switchX_dpid = v
            switchY_dpid = path[k + 1]  # Next switch in path
            (switchX_out_port, switchY_in_port) = self.neighbor_ports[(
                switchX_dpid, switchY_dpid)]

            switchX_datapath = self.switches[v]

            self._add_switch_flow(
                switchX_datapath, in_port, hostA['mac'],
                hostZ['mac'], switchX_out_port)

            in_port = switchY_in_port

    def _add_switch_flow(self, datapath, in_port, eth_src, eth_dst, out_port):
        parser = datapath.ofproto_parser
        priority = 100
        match = parser.OFPMatch(
            in_port=in_port,
            eth_src=eth_src,
            eth_dst=eth_dst
        )
        actions = [parser.OFPActionOutput(out_port)]
        self.logger.info('%8d %8d %17s %17s %8d %8d', datapath.id,
                         in_port, eth_src, eth_dst, out_port, priority)
        self.add_flow(datapath, priority, match, actions)

    # Taken from Ryu book page 88:
    def _send_packet(self, datapath, port, pkt):
        ofproto = datapath.ofproto
        parser = datapath.ofproto_parser
        pkt.serialize()
        self.logger.debug("packet-out %s" % (pkt,))
        data = pkt.data
        actions = [parser.OFPActionOutput(port=port)]
        out = parser.OFPPacketOut(datapath=datapath,
                                  buffer_id=ofproto.OFP_NO_BUFFER,
                                  in_port=ofproto.OFPP_CONTROLLER,
                                  actions=actions, data=data)
        datapath.send_msg(out)

    # Inspired by Ryu book page 88:
    def _send_arp_reply(self, datapath, port, pkt_ethernet, pkt_arp, src_mac, src_ipv4):
        if pkt_arp.opcode != arp.ARP_REQUEST:
            return
        pkt = packet.Packet()
        pkt.add_protocol(ethernet.ethernet(
            ethertype=pkt_ethernet.ethertype,
            dst=pkt_ethernet.src, src=src_mac))
        pkt.add_protocol(arp.arp(opcode=arp.ARP_REPLY, src_mac=src_mac,
                                 src_ip=src_ipv4, dst_mac=pkt_arp.src_mac,
                                 dst_ip=pkt_arp.src_ip))
        self._send_packet(datapath, port, pkt)

    def _handle_arp_request(self, datapath, port, pkt_ethernet, pkt_arp, src_mac, src_ipv4):
        ofproto = datapath.ofproto
        parser = datapath.ofproto_parser
        match = parser.OFPMatch(in_port=port, eth_src=pkt_ethernet.src, eth_dst='FF:FF:FF:FF:FF:FF')
        #match = parser.OFPMatch(in_port=port)
        inst = [parser.OFPInstructionActions(
                ofproto.OFPIT_CLEAR_ACTIONS, [])]
        mod = parser.OFPFlowMod(datapath=datapath, priority=65535,
                                match=match, instructions=inst, hard_timeout=60)
        datapath.send_msg(mod)  # FIXME: I don't get any hits
        self._send_arp_reply(datapath, port, pkt_ethernet, pkt_arp, src_mac, src_ipv4)

    def _handle_new_host(self, datapath, switch_port, host_mac):
        ofproto = datapath.ofproto
        parser = datapath.ofproto_parser
        match = parser.OFPMatch(in_port=switch_port, eth_src=host_mac, eth_dst='FF:FF:FF:FF:FF:FF')
        inst = [parser.OFPInstructionActions(
                ofproto.OFPIT_CLEAR_ACTIONS, [])]
        mod = parser.OFPFlowMod(datapath=datapath, priority=65535,
                                match=match, instructions=inst, hard_timeout=5)
        datapath.send_msg(mod)

        parser = datapath.ofproto_parser
        match = parser.OFPMatch(eth_dst=host_mac)
        actions = [parser.OFPActionOutput(switch_port)]
        self.add_flow(datapath, 50, match, actions)

    @set_ev_cls(ofp_event.EventOFPPacketIn, MAIN_DISPATCHER)
    def _packet_in_handler(self, ev):
        # If you hit this you might want to increase
        # the "miss_send_length" of your switch
        if ev.msg.msg_len < ev.msg.total_len:
            self.logger.debug("packet truncated: only %s of %s bytes",
                              ev.msg.msg_len, ev.msg.total_len)
        msg = ev.msg
        datapath = msg.datapath
        ofproto = datapath.ofproto
        parser = datapath.ofproto_parser
        in_port = msg.match['in_port']

        pkt = packet.Packet(msg.data)
        eth = pkt.get_protocols(ethernet.ethernet)[0]

        if eth.ethertype == ether_types.ETH_TYPE_LLDP:
            # ignore lldp packet
            return

        print 'PACKET IN >>', datapath.id, in_port
        self.logger.debug("packet-in %s" % (pkt,))

        dst = eth.dst
        src = eth.src

        dpid = datapath.id

        # self.hosts = {'10.0.0.1': {'mac': '00:00:00:00:00:01', 'switch': 11, 'port': 1},
        #              '10.0.0.2': {'mac': '00:00:00:00:00:02', 'switch': 13, 'port': 2}}

        # Check if this is a Layer2 broadcast:
        if eth.dst == 'ff:ff:ff:ff:ff:ff':
            # Check if this is an ARP Request.
            # ARP Responses are unicast.
            # Inspired by Ryu book page 88
            pkt_arp = pkt.get_protocol(arp.arp)
            if pkt_arp:
                if pkt_arp.opcode != arp.ARP_REQUEST:
                    print 'Expected an ARP Request but this is not it.'
                    return
                # Learn new source host:
                if pkt_arp.src_ip not in self.hosts:
                    self.hosts[pkt_arp.src_ip] = {
                        'mac': src, 'switch': dpid, 'port': in_port}
                    print "DISCOVERED NEW HOST", pkt_arp.src_ip, self.hosts[pkt_arp.src_ip]
                    # Add flow on this switch, point to NEW HOST:
                    self._handle_new_host(datapath, in_port, src)
                    return

                # Run SPF and create end-to-end path
                if pkt_arp.dst_ip in self.hosts:
                    dst_switch = self.hosts[pkt_arp.dst_ip]['switch']

                    self._handle_arp_request(
                        datapath, in_port, eth, pkt_arp,
                        self.hosts[pkt_arp.dst_ip]['mac'], pkt_arp.dst_ip)

                    current_time = int(time.time())  # 2017-10-17
                    if pkt_arp.src_ip not in self.hosts_temp_ban:  # 2017-10-17
                        self.hosts_temp_ban[pkt_arp.src_ip] = current_time
                        print "BANNED HOST >>", pkt_arp.src_ip, "FROM ARP REQ FOR 10sec"
                    elif current_time - self.hosts_temp_ban[pkt_arp.src_ip] < 10:
                        print "HOST", pkt_arp.src_ip, "BANNED FROM ARP"
                        return


                    all_paths = self._find_paths(self.topo_links, dpid, dst_switch)
                    if all_paths is None:
                        # Hosts are on same switch. I don't need to build
                        # a flow. It was already done by _handle_new_host()
                        return

                    hop_count = self._calculate_hop_count(all_paths)
                    nfm_metric = self._calculate_path_metric(all_paths)

                    # Weights for composite metric:
                    w1 = 0.1  # Weight for Hop Count
                    w2 = 0.9  # Weight for NFM metric
                    best_metric = 100000000000  # Just an Arbitrary high value.
                    # Lowest metric means best metric! (0 is the best possible)
                    best_path = []
                    for path in all_paths:
                        composite_metric = w1 * hop_count[path]
                        composite_metric += w2 * nfm_metric[path]
                        if composite_metric < best_metric:
                            best_metric = composite_metric
                            best_path = path

                    print 'BEST PATH between', (dpid, dst_switch), 'is', best_path
                    print 'with Composite metric', best_metric


                    best_path = list(best_path)  # Change to list, 2017-10-19
                    # Installing flows from A >> Z
                    self._add_end_to_end_flow(
                        best_path, pkt_arp.src_ip, pkt_arp.dst_ip)
                return

        self.logger.info("TABLE MISS: switch %s, packet-in %s", dpid, (pkt,))
