# ~/ryu$ PYTHONPATH= ./bin/ryu-manager ryu/app/cpm_13.py --observe-links

from ryu.ofproto import ofproto_v1_3
from ryu.app import simple_switch_13
from ryu.controller.handler import set_ev_cls
from ryu.topology import event, switches
from ryu.topology.api import get_switch, get_link
from ryu.controller import ofp_event
from ryu.controller.handler import CONFIG_DISPATCHER, MAIN_DISPATCHER
from ryu.lib.packet import packet
from ryu.lib.packet import arp
from ryu.lib.packet import ethernet
from ryu.lib.packet import ether_types
from ryu.lib.packet import ipv4
from ryu.lib.packet import icmp
from ryu.lib.packet import udp
from ryu.lib.packet import tcp
from ryu.lib import hub
import time
from my_http_client import MyHTTPClient
from cpm_13_lib import SwitchPair


class CPM13(simple_switch_13.SimpleSwitch13):
    "CPM module for topology discovery and packet forwarding"

    OFP_VERSIONS = [ofproto_v1_3.OFP_VERSION]

    def __init__(self, *args, **kwargs):
        super(CPM13, self).__init__(*args, **kwargs)
        self.rib = {}  # Contains all paths
        self.fib = {}  # Contains only paths used for packet forwarding

        self.topo_links = {}
        self.switches = {}
        self.neighbor_ports = {}  # Temp 2017-10-02
        self.port_metric = {}  # 2017-10-16
        self.hosts = {}
        self.sample_interval = 15  # How often to query the DM
        self.ref_bw = 100000000.0  # 100 Mbps; Reference Bandwidth in Bps
        self.ref_hc = 30.0  # 30 Hops; Reference Hop Count
        # Connect with DM:
        self.dm_server = MyHTTPClient(
            'http://192.16.125.183/API_REST/request.php/get?')
        self.port_counters_snapshot = {'object_index': 1000, 'switches': []}
        hub.spawn(self._get_dm_data)

        # Connect with DM#2; need a second object of MyHTTPClient as
        # URL for POST requests is different.
        self.dm_server_2 = MyHTTPClient(
            'http://192.16.125.183/API_REST/request.php')
        reset_topo_link_json_obj = {'src_switch': '',
                                    'src_sw_port': '',
                                    'dst_switch': '',
                                    'dst_sw_port': '',
                                    'module': 'topology',
                                    'event': 'new_topology',
                                    'bandwidth': ''}
        self.dm_server_2.send_POST(reset_topo_link_json_obj)

    def _get_dm_data(self):
        while True:
            # Ask for DM data every self.sample_interval seconds:
            hub.sleep(self.sample_interval)
            #body = self.dm_server.get_data_from_dm('port_counters', 1)
            params_dict = {'module': 'nfm', 'row': 1, 'table': 'port_counters'}
            body = self.dm_server.send_GET(params_dict)
            # print body
            if body is None:  # This is True if NO json in HTTP reponse
                print 'No NFM monitoring data received from DM'
                continue
            # print body
            self._calculate_link_metric(body)
            self._rib_update()

    def _rib_update(self):
        # Update the metric for all paths in RIB
        for (a, z), pair in self.rib.items():
                all_paths = pair.getAllPaths().keys()
                for path in all_paths:
                    metric = self._calculate_path_uni_metric(path, 3)
                    pair.updateSingleMetric(path, metric)
                    rv_path = path[::-1]
                    rv_metric = self._calculate_path_uni_metric(rv_path, 3)
                    pair.updateSingleMetric(rv_path, rv_metric)
                    time.sleep(0.00002)  # sleep thread for 20ns

        # Also update FIB
        self._fib_update()

    def _rib_lookup(self, switches_tuple):
        # switches is a tuple with two elements
        # Fuction returns the corresponding SwitchPair object
        if switches_tuple in self.rib.keys():
            pair = self.rib[switches_tuple]
        elif (switches_tuple[1], switches_tuple[0]) in self.rib.keys():
            pair = self.rib[(switches_tuple[1], switches_tuple[0])]
        else:
            pair = None
            print '[FIB UPDATE] Something went wrong; pair not in RIB'

        return pair

    def _fib_update(self):
        for (a, z) in self.fib.keys():
            pair = self._rib_lookup((a, z))
            if pair is None:
                print '[FIB UPDATE] Something went wrong; pair not in RIB'

            (best_path, best_metric) = pair.getBestPath(a, z)
            if self.fib[(a, z)] != best_path:
                self.fib[(a, z)] = best_path
                print '[FIB] updated ', (a, z), self.fib[(a, z)], best_metric

    def _fib_lookup(self, src_switch, dst_switch):

        # Not in FIB
        if (src_switch, dst_switch) not in self.fib:
            pair = self._rib_lookup((src_switch, dst_switch))
            (best_path, best_metric) = pair.getBestPath(src_switch, dst_switch)
            print '[FIB] added', (src_switch, dst_switch), best_path, best_metric 
            self.fib[(src_switch, dst_switch)] = best_path

        return self.fib[(src_switch, dst_switch)]

    def _fib_next_hop(self, current_switch, src_ip, dst_ip):
        # There should be no situation where current_switch
        # is the last switch of a path. This is due to the flow
        # installed by handle_new_host()

        a = self.hosts[src_ip]['switch']
        z = self.hosts[dst_ip]['switch']
        path = self._fib_lookup(a, z)

        if current_switch in path:
            next_switch = path.index(current_switch) + 1
            (out_port, _) = self.neighbor_ports[(current_switch,
                                                 path[next_switch])]
            print '[SWITCH]', current_switch, ': Packet', src_ip, dst_ip, 'Out port', out_port
            return out_port
        else:
            print '[SWITCH]', current_switch, 'cannot FWD', src_ip, dst_ip
            return None

    def _add_switch_flow(self, datapath, in_port, pkt, buffer_id=None):
        parser = datapath.ofproto_parser

        match_args = self._parse_ip_packet_headers(pkt)

        match_args['in_port'] = in_port

        if match_args['ip_proto'] == 1:
            # ICMP
            priority = 200

        elif match_args['ip_proto'] == 6:
            # TCP
            # Add src and dst port
            priority = 400

        elif match_args['ip_proto'] == 17:
            # UDP
            # Add src and dst port
            priority = 300

        else:
            priority = 100  # Priority for everything else

        out_port = self._fib_next_hop(datapath.id,
                                      match_args['ipv4_src'],
                                      match_args['ipv4_dst'])

        # https://docs.python.org/2/tutorial/controlflow.html#unpacking-argument-lists
        match = parser.OFPMatch(**match_args)
        self.logger.info('%8d %8d %17s %17s %17s %17s %8d %8d %8d',
                         datapath.id,
                         in_port,
                         match_args['eth_src'], match_args['ipv4_src'],
                         match_args['eth_dst'], match_args['ipv4_dst'],
                         out_port,
                         match_args['ip_proto'], priority)

        actions = [parser.OFPActionOutput(out_port)]
        self.add_flow(datapath, priority, match, actions, buffer_id)

    def _calculate_link_metric(self, data_dict):

        for switch in data_dict['switches']:
            # each element is a dictionary as well
            for port_counter in switch['port_counters']:
                # each element is a dictionary as well

                tx_metric = float(port_counter['tx_bw']) / self.ref_bw

                # Round to 3 decimal points
                # https://docs.python.org/2.7/library/functions.html#round
                tx_metric = round(tx_metric, 3)

                self.port_metric[(switch['switch_dpid'],
                                  port_counter['port'])] = tx_metric

                # print '[CALC] Port Metric', (switch['switch_dpid'], port_counter['port']), 'set to', tx_metric

    def _calculate_path_uni_metric(self, path, metric_type=1):

        # Hop Count
        hop_count = len(path)
        hop_metric = hop_count / self.ref_hc

        if metric_type == 1:
            # Return Hop Metric
            return hop_metric

        # NFM metric
        # Find the busiest port in the path.
        # Path metric is equal to the busiest port's metric.
        worst_metric = 0
        for k, v in enumerate(path[:-1]):
            # v is the current switch dpid
            # path[k + 1] is the dpid of the next switch in path
            (tx_port, _) = self.neighbor_ports[(v, path[k + 1])]
            # Find busiest port in path:
            if worst_metric < self.port_metric[(v, tx_port)]:
                worst_metric = self.port_metric[(v, tx_port)]

        if metric_type == 2:
            # Return NFM Metric
            return worst_metric

        # Composite (Hop Count and NFM metric)
        # Weights for composite metric:
        w1 = 0.1  # Weight for Hop Count
        w2 = 0.9  # Weight for NFM metric
        # Lowest metric means best metric! (0 is the best possible)
        composite_metric = w1 * hop_count + w2 * worst_metric

        if metric_type == 3:
            # Return Composite metric
            return composite_metric

    def _deny_flow(self, datapath, in_port, pkt_in, hard_timeout=1):
        ofproto = datapath.ofproto
        parser = datapath.ofproto_parser

        match_args = self._parse_ip_packet_headers(pkt_in)
        match_args['in_port'] = in_port

        # https://docs.python.org/2/tutorial/controlflow.html#unpacking-argument-lists
        match = parser.OFPMatch(**match_args)

        inst = [parser.OFPInstructionActions(
                ofproto.OFPIT_CLEAR_ACTIONS, [])]
        mod = parser.OFPFlowMod(datapath=datapath, priority=65535,
                                match=match, instructions=inst,
                                hard_timeout=hard_timeout)
        datapath.send_msg(mod)

    @set_ev_cls(event.EventSwitchEnter)
    def _handle_new_switch(self, ev):
        # Gives the ryu.controller.controller.Datapath object of the switch:
        # print ev.switch.dp
        new_switch = ev.switch.dp
        self.switches[new_switch.id] = ev.switch.dp
        self.topo_links[new_switch.id] = []

    @set_ev_cls(event.EventLinkAdd)
    def _handle_new_link(self, ev):
        # The link obviously connects two switches. The following two:
        # SwitchA: ev.link.src
        # SwitchB: ev.link.dst
        # are Port objects (switches.py)
        # print 'New link between', ev.link.src.dpid, 'and', ev.link.dst.dpid

        switchA = ev.link.src
        switchB = ev.link.dst

        # print switchA

        self.topo_links[switchA.dpid].append(switchB.dpid)
        # I don't need to report the other direction.
        # SwitchB will also raise this event.

        # {11: [12, 21], 12: [11, 13], 13: [22, 12], 22: [21, 13],
        # 21: [22, 11]}

        # print 'Neighbors', self.topo_links

        # Temp 2017-10-02
        self.neighbor_ports[(switchA.dpid, switchB.dpid)] = (
            switchA.port_no, switchB.port_no)
        # print 'Ports', self.neighbor_ports

        # 2017-10-16
        self.port_metric[(switchA.dpid, switchA.port_no)] = 0
        # print 'Port Metric', (switchA.dpid, switchA.port_no), 'set to', 0

        # 2017-11-16
        # Inform DM about this link
        new_topo_link_json_obj = {'src_switch': switchA.dpid,
                                  'src_sw_port': switchA.port_no,
                                  'dst_switch': switchB.dpid,
                                  'dst_sw_port': switchB.port_no,
                                  'module': 'topology',
                                  'event': 'link_up',
                                  'bandwidth': 100000000}
        self.dm_server_2.send_POST(new_topo_link_json_obj)

    # https://www.python.org/doc/essays/graphs/
    def _find_all_paths(self, neighbors, start, end, path=[]):
        path = path + [start]
        if start == end:
            return [path]
        if start not in neighbors.keys():
            return []
        paths = []
        for node in neighbors[start]:
            if node not in path:
                newpaths = self._find_all_paths(neighbors, node, end, path)
                for newpath in newpaths:
                    paths.append(newpath)
        return paths

    def _parse_ip_packet_headers(self, pkt):
        # The following headers are available to all IPv4 packets
        pkt_ethernet = pkt.get_protocol(ethernet.ethernet)
        pkt_ipv4 = pkt.get_protocol(ipv4.ipv4)
        eth_type = pkt_ethernet.ethertype
        eth_src = pkt_ethernet.src
        ip_src = pkt_ipv4.src
        eth_dst = pkt_ethernet.dst
        ip_dst = pkt_ipv4.dst
        ip_proto = pkt_ipv4.proto  # 1 for ICMP, 6 for TCP, 17 for UDP

        match_args = {
            'eth_type': eth_type,
            'eth_src': eth_src,
            'ipv4_src': ip_src,
            'eth_dst': eth_dst,
            'ipv4_dst': ip_dst,
            'ip_proto': ip_proto
        }

        if ip_proto == 1:
            # ICMP
            # No other match argument needed
            pass

        elif ip_proto == 6:
            # TCP
            # Add src and dst port
            pkt_tcp = pkt.get_protocol(tcp.tcp)
            match_args['tcp_src'] = pkt_tcp.src_port
            match_args['tcp_dst'] = pkt_tcp.dst_port

        elif ip_proto == 17:
            # UDP
            # Add src and dst port
            pkt_udp = pkt.get_protocol(udp.udp)
            match_args['udp_src'] = pkt_udp.src_port
            match_args['udp_dst'] = pkt_udp.dst_port

        else:
            # Everything else
            pass

        return match_args

    # Taken from Ryu book page 88:
    def _send_packet(self, datapath, port, pkt):
        ofproto = datapath.ofproto
        parser = datapath.ofproto_parser
        pkt.serialize()
        self.logger.debug("packet-out %s" % (pkt,))
        data = pkt.data
        actions = [parser.OFPActionOutput(port=port)]
        out = parser.OFPPacketOut(datapath=datapath,
                                  buffer_id=ofproto.OFP_NO_BUFFER,
                                  in_port=ofproto.OFPP_CONTROLLER,
                                  actions=actions, data=data)
        datapath.send_msg(out)

    # Inspired by Ryu book page 88:
    def _handle_arp_request(self, datapath, in_port, pkt_in):
        pkt_in_ether = pkt_in.get_protocol(ethernet.ethernet)
        pkt_in_arp = pkt_in.get_protocol(arp.arp)

        if pkt_in_arp.opcode != arp.ARP_REQUEST:
            return
        elif pkt_in_arp.dst_ip not in self.hosts:
            print 'Unknown host. Cannot send ARP Reply.'
            return

        src_mac = self.hosts[pkt_in_arp.dst_ip]['mac']

        pkt_out = packet.Packet()
        pkt_out.add_protocol(ethernet.ethernet(
            ethertype=pkt_in_ether.ethertype,
            dst=pkt_in_ether.src, src=src_mac))
        pkt_out.add_protocol(arp.arp(opcode=arp.ARP_REPLY,
                                     src_mac=src_mac,
                                     src_ip=pkt_in_arp.dst_ip,
                                     dst_mac=pkt_in_arp.src_mac,
                                     dst_ip=pkt_in_arp.src_ip))
        self._send_packet(datapath, in_port, pkt_out)

    def _rib_init(self, new_switch):
        switches_with_hosts = []
        for ip, params in self.hosts.items():
            if params['switch'] not in switches_with_hosts:
                switches_with_hosts.append(params['switch'])
        for switch in switches_with_hosts:
            if switch == new_switch:
                # Same switch, no need to add to RIB
                continue
            if (new_switch, switch) not in self.rib and (switch, new_switch) not in self.rib:
                new_pair = SwitchPair()
                new_pair.importAllPaths(self._find_all_paths(
                    self.topo_links, new_switch, switch))
                self.rib[(new_switch, switch)] = new_pair
                print '[RIB] Initialized for', (new_switch, switch), ' and ', (switch, new_switch)
            else:
                print 'RIB Initialization failed', self.rib

    def _handle_new_host(self, datapath, switch_port, host_mac, host_ip):
        # Add new host to host dict
        self.hosts[host_ip] = {'mac': host_mac,
                               'switch': datapath.id,
                               'port': switch_port}

        print "[DISCOVERED NEW HOST]", host_ip, self.hosts[host_ip]

        # Add flow on this switch, point to NEW HOST:
        parser = datapath.ofproto_parser
        match = parser.OFPMatch(eth_dst=host_mac)
        actions = [parser.OFPActionOutput(switch_port)]
        self.add_flow(datapath, 50, match, actions)

        # Initialize RIB
        self._rib_init(datapath.id)

    @set_ev_cls(ofp_event.EventOFPPacketIn, MAIN_DISPATCHER)
    def _packet_in_handler(self, ev):
        # If you hit this you might want to increase
        # the "miss_send_length" of your switch
        if ev.msg.msg_len < ev.msg.total_len:
            self.logger.debug("packet truncated: only %s of %s bytes",
                              ev.msg.msg_len, ev.msg.total_len)

        msg = ev.msg
        datapath = msg.datapath
        in_port = msg.match['in_port']

        pkt = packet.Packet(msg.data)
        eth = pkt.get_protocols(ethernet.ethernet)[0]

        if eth.ethertype == ether_types.ETH_TYPE_LLDP:
            # ignore lldp packet
            return

        # print 'PACKET IN >>', datapath.id, in_port
        self.logger.debug("packet-in %s" % (pkt,))

        dpid = datapath.id

        # self.hosts = {'10.0.0.1': {'mac': '00:00:00:00:00:01', 'switch': 11, 'port': 1},
        #              '10.0.0.2': {'mac': '00:00:00:00:00:02', 'switch': 13, 'port': 2}}

        # Handle ARP Request packets:
        pkt_arp = pkt.get_protocol(arp.arp)
        if pkt_arp:
            # Inspired by Ryu book page 88
            if pkt_arp:
                if pkt_arp.opcode != arp.ARP_REQUEST:
                    print 'Expected an ARP Request but this is not it.'
                    return

                # Learn new source host:
                if pkt_arp.src_ip not in self.hosts:
                    self._handle_new_host(
                        datapath, in_port, eth.src, pkt_arp.src_ip)
                    return

                # Reply to ARP request:
                if pkt_arp.dst_ip in self.hosts:
                    self._handle_arp_request(datapath, in_port, pkt)
                return

        # Handle IPv4 packets:
        pkt_ipv4 = pkt.get_protocol(ipv4.ipv4)
        if not pkt_ipv4:
            print 'NOT an IPv4 Packet:'
            self.logger.info(
                "TABLE MISS: switch %s, packet-in %s", dpid, (pkt,))
            return
        else:
            # Block all similar future packets for 1 second:
            # self._deny_flow(datapath, in_port, pkt)

            # dst_switch = self.hosts[pkt_ipv4.dst]['switch']
            # Find best path between (src_switch, dst_switch)
            # best_path = self._find_best_path(dpid, dst_switch)
            # print datapath.id, in_port, pkt_ipv4.src, pkt_ipv4.dst
            # best_path = self._fib_lookup(dpid, dst_switch)

            # Installing flows from A >> Z
            # self._add_network_uni_flow(best_path, pkt, msg.buffer_id)
            self._add_switch_flow(datapath, in_port, pkt, msg.buffer_id)
